#include "interpreter.h"
#include "config.h"
#include "data.h"


///////////////////////////////////////////////////////////////////////////////


Interpreter::Interpreter(CCOsd* c) : cThread("CC Interpreter")
{
  dprint(L_DBG, "Interpr.: Constructor.");
  inBuffer = new cRingBufferLinear(KILOBYTE(32), 32, false, "CC Interpr. In");
  osd = c;
}


//-----------------------------------------------------------------------------

Interpreter::~Interpreter() 
{
  Cancel(3);
  
  delete inBuffer;
  dprint(L_DBG, "Interpr.: Destructor.");
}


//-----------------------------------------------------------------------------

void Interpreter::Put(uint8_t* data, int length)
{
  // Strip MSB
  for (int i=0; i<length; i++)
    data[i] = data[i] & 0x7F;
    
  int p = inBuffer->Put(data, length);
  if (p != length) { 
    dprint(L_ERR, "Interpr.: Input buffer overflow (%d bytes dropped)", length - p); 
    inBuffer->ReportOverflow(length - p);
  }
}


//-----------------------------------------------------------------------------

void Interpreter::Action(void)
{
  dprint(L_DBG, "Interpr.: Thread Started.");
  
  while (Running())
  {
    //uint64_t st = cTimeMs::Now();
    
    int length = 0;
    int used = 0;
    uint8_t* data = inBuffer->Get(length);
    if (data && length != 0)
    {
#if 0          
      fprintf(stderr, "\n*****************************\n");
      for (int i=0; i<length; i++)
        //fprintf(stderr, "%c", data[i] & 0x7F);
        fprintf(stderr, "0x%02X ", data[i] & 0x7F);
      fprintf(stderr, "\n*****************************\n\n");
#endif
      uint8_t* d = data;    
      while (d-data < length)
      {       
        uint8_t* start = NULL;
        while (d-data < length) {
          if (d[0] != 0x00) {
            start = d;
            break;
          }
          d++;
        }  
        if (!start) {
          used = length;
          break;
        } 
        
        //dprint(L_DBG, "Interpr.: Found start at %d.", start-data);
        
        uint8_t* end = NULL;
        while (d-data < length) {
          if (d[0] == 0x00) {
            end = d;
            break;
          }
          d++;
        }  
        if (!end) {
          used = start - data;
          //dprint(L_DBG, "Interpr.: No end found, used %d of %d", used, length);
          break;
        } 
        
        // Always feed Process an even number of bytes
        Process(start, end-start + (end-start)%2);
      }
      
      inBuffer->Del(used);
    }
    //uint64_t tst = SLEEP_TIME - cTimeMs::Now() + st;
    //dprint(L_DBG, "Interpr.: Sleeping for %d ms.", tst);
    cCondWait::SleepMs(SLEEP_TIME);
  }
  
  dprint(L_DBG, "Interpr.: Thread Ended."); 
}


//-----------------------------------------------------------------------------

uint8_t GetStdChar(uint8_t c)
{
  uint8_t r = c;
  switch (c)
  {
    case 0x2A: r = 0xE1; break;	// lower-case a, acute accent
    case 0x5C: r = 0xE9; break;	// lower-case e, acute accent
    case 0x5E: r = 0xED; break;	// lower-case i, acute accent
    case 0x5F: r = 0xF3; break;	// lower-case o, acute accent
    case 0x60: r = 0xFA; break;	// lower-case u, acute accent
    case 0x7B: r = 0xE7; break;	// lower-case c with cedilla
    case 0x7C: r = 0xF7; break;	// division symbol
    case 0x7D: r = 0xD1; break;	// upper-case enya (N-tilde)
    case 0x7E: r = 0xF1; break;	// lower-case enya (n-tilde)
    case 0x7F: r = 0x90; break;	// solid block (using hollow box)
  }
  
  return r;
}



void Interpreter::Process(uint8_t* data, int length)
{  
#if 0                      
  fprintf(stderr, "\n=============================\n");
  for (int i=0; i<length; i++)
    //fprintf(stderr, "%c", data[i] & 0x7F);
    fprintf(stderr, "0x%02X ", data[i] & 0x7F);
  fprintf(stderr, "\n=============================\n\n");
  
  return;
#endif 
 
  dprint(L_TXT, "\n=============================\n");  
  for(int i=0; i<length-1; i+=2) // Should start at 0 or 2 ?
  {
    uint8_t hi = data[i];
    uint8_t lo = data[i+1];
    uint16_t code = (hi << 8) | lo;
    
    // ~~~~ Standard Character ~~~~
    if (hi >= 0x20 )//|| hi == 0x00)
    {
      dprint(L_TXT, "%c%c", hi, lo);
      osd->AddChar( GetStdChar(hi) ); 
      if (lo != 0x00) 
        osd->AddChar( GetStdChar(lo) ); 
        
      continue;  
    }
    
    // Check channel
    uint8_t chan = (hi & 0x08) >> 3;
    if (config.CCType != chan) {
      // dprint(L_DBG, "Interpr.: Want ch. %d but got ch. %d.", config.CCType, chan);
      return;
    }
      
      
    // ~~~~ Control Code 1 ~~~~
    if ( ((hi == 0x14) || (hi == 0x1C) || (hi == 0x15) || (hi == 0x1D)) && 
              ((lo >= 0x20) && (lo <= 0x2F))) 
    { 
      uint8_t cntrlCode = lo & 0xF;
      dprint(L_TXT, "%s", cntrlCodes[cntrlCode]);
  
      switch (cntrlCode)
      {
        case CMD_RCL: // Resume Caption Loading
          osd->SetMode(M_RCL);
          break;
        case CMD_BS:  // Backspace
          osd->Backspace();
          break;
        case CMD_AOF: // Reserved
        case CMD_AON:
          break;
        case CMD_DER: // Delete to end of row
          osd->DER();
          break;
        case CMD_RU2: // Roll-up captions, 2 rows 
        case CMD_RU3: // Roll-up captions, 3 rows
        case CMD_RU4: // Roll-up captions, 4 rows
          if ( osd->SetMode(M_RU, 2+cntrlCode-CMD_RU2) )
          {
            osd->ClearScreen();
            osd->ClearBuffer();
            osd->SetRow(15);
          }
          break;          
        case CMD_FON: // Flash on
          break;                                        
        case CMD_RDC: // Resume direct captioning
          osd->SetMode(M_RDC);
          break;
        case CMD_TR:  // Text restart
          break;            
        case CMD_RTD: // Resume text display
          break;
        case CMD_EDM: // Erase displayed memory
          osd->ClearScreen();
          break;
        case CMD_CR:  // Carriage return
          osd->CR();
          break;
        case CMD_ENM: // Erase nondisplayed memory 
          osd->ClearBuffer();   
          break;                     
        case CMD_EOC: // End of Caption
          osd->Flush();
          break;  
      }
    }
    
    // ~~~~ Control Code 2 ~~~~
    else if ( ((hi == 0x17) || (hi == 0x1F)) && ((lo >= 0x21) && (lo <= 0x23))) 
    {
      uint8_t cntrlCode = ((lo-0x21) & 0x3)+0x10;
      dprint(L_TXT, "%s", cntrlCodes[cntrlCode]);
      osd->TO(cntrlCode - CMD_TO1 + 1);
    }
    
    //  ~~~~Mid-Row Code ~~~~
    else if ( ((hi == 0x11) || (hi == 0x19)) && ((lo >= 0x20) && (lo <= 0x2F))) 
    {
      osd->AddChar(0x20); // Add space
      dprint(L_TXT, "{MRC %s %s}", midRowCode[(lo >> 1) & 0x07],
                                   underlineCode[lo & 0x01]);
    }
    
    // ~~~~ Attribute Codes ~~~~
    else if ( ((hi == 0x10) || (hi == 0x18)) && ((lo >= 0x20) && (lo <= 0x27))) 
    {
      // TYPE_ATTRIBUTE_1;
      dprint(L_TXT, "[A1]");
    }
    else if ( (code == 0x172D) || (code == 0x1F2D)) 
    {
      // TYPE_ATTRIBUTE_2;
      dprint(L_TXT, "[A2]");
    }
    else if ( (code==0x172E) || (code==0x172F) || (code==0x1F2E) || (code==0x1F2F) ) 
    {
      // TYPE_ATTRIBUTE_3;
      dprint(L_TXT, "[A3]");
    }

    // ~~~~ Optional Closed Group Extensions ~~~~
    else if ( ((hi == 0x17) || (hi == 0x1F)) && ((lo >= 0x24) && (lo <= 0x2A))) 
    {
      // TYPE_CLOSED_GROUP_EXT;
      dprint(L_TXT, "[CGE]");
    }

    // ~~~~ Standard extended characters ~~~~
    else if ( ((hi == 0x11) || (hi == 0x19)) &&  ((lo >= 0x30) && (lo <= 0x3F))) 
    {
      //dprint(L_TXT, "[E1: %c]", ext1Char[lo-0x30]);
      osd->AddChar(ext1Char[lo-0x30]);
    }

    // ~~~~ 2nd Extended Character set ~~~~
    else if ( ((hi == 0x12) || (hi == 0x1A)) && ((lo >= 0x20) && (lo <= 0x3F))) 
    {
      dprint(L_TXT, "[E2(%d): %c]", lo-0x20, ext2Char[lo-0x20]);
      osd->AddChar(ext2Char[lo-0x20]);
    }

    // ~~~~ 3rd Extended Character set ~~~~
    else if ( ((hi == 0x13) || (hi == 0x1B)) && ((lo >= 0x20) && (lo <= 0x3F))) 
    {
      dprint(L_TXT, "[E3: %c]", ext3Char[lo-0x20]);
      osd->AddChar(ext3Char[lo-0x20]);
    }

    // ~~~~ Extended Data service ~~~~
    else if ((hi >= 0x01) && (hi <= 0x0F)) 
    {
      // TYPE_XDS;
      dprint(L_TXT, "[XDS]");
    }
    
    // ~~~~ Check for PreAmble Address Codes ~~~~
    else 
    {
      // Strip out Channel, Colors, Underline
      uint16_t controlCode1 = code & 0xF760;
      int row = 0;
      int col = 0;  
      switch (controlCode1) {
        case 0x1140: row =  1; break;
        case 0x1160: row =  2; break;
        case 0x1240: row =  3; break;
        case 0x1260: row =  4; break;
        case 0x1540: row =  5; break;
        case 0x1560: row =  6; break;
        case 0x1640: row =  7; break;
        case 0x1660: row =  8; break;
        case 0x1740: row =  9; break;
        case 0x1760: row = 10; break;
        case 0x1040: row = 11; break;
        case 0x1340: row = 12; break;
        case 0x1360: row = 13; break;
        case 0x1440: row = 14; break;
        case 0x1460: row = 15; break; 
      }
      
      uint8_t pCode = (lo >> 1) & 0x0F;
      switch (pCode)
      {
        case 0x00: // White
        case 0x01: // Green
        case 0x02: // Blue
        case 0x03: // Cyan
        case 0x04: // Red
        case 0x05: // Yellow
        case 0x06: // Magenta
        case 0x07: // Italics
          break;
        case 0x08: col =  0; break; // Indent 0, White
        case 0x09: col =  4; break; // Indent 4, White
        case 0x0A: col =  8; break; // Indent 8, White
        case 0x0B: col = 12; break; // Indent 12, White
        case 0x0C: col = 16; break; // Indent 16, White
        case 0x0D: col = 20; break; // Indent 20, White
        case 0x0E: col = 24; break; // Indent 24, White
        case 0x0F: col = 28; break; // Indent 28, White
      }
      
      if (row) { //Got PAC
        dprint(L_TXT, "{%s,(%d,%d)}", preambleCode[pCode], col,row);
        osd->SetPosition(row-1, col);
      }
      else {
        dprint(L_TXT, "[UT]"); 
      }
    }
    
          
    // Check for duplicates
    if (i+2 < length-1) 
    {
      if ( (hi == data[i+2]) && (lo == data[i+3]) ) 
      { i+=2; }   
    }
  }
  
  dprint(L_TXT, "\n=============================\n\n");       
}


///////////////////////////////////////////////////////////////////////////////

