#include "osd.h"
#include "common.h"
#include "config.h"

///////////////////////////////////////////////////////////////////////////////


#define SCREENWIDTH  720
#define SCRRENHEIGHT 480
#define OSDWIDTH  576 
#define OSDHEIGHT 384
#define NUMROW 16
#define NUMCOL 32
#define CELLWIDTH (OSDWIDTH/NUMCOL)
#define CELLHEIGHT (OSDHEIGHT/NUMROW)

#define STARTX ((SCREENWIDTH-OSDWIDTH)/2)
#define STARTY ((SCRRENHEIGHT-OSDHEIGHT)/2)


///////////////////////////////////////////////////////////////////////////////


CCOsd::CCOsd()
{
  osd = NULL;
  offScreenBuffer = new cBitmap(OSDWIDTH, OSDHEIGHT, 2);
  scroller = NULL;
  
  osdMutex = new cMutex();
  scrollerMutex = new cMutex();
  bufferMutex = new cMutex();
  
  font = cFont::GetFont(fontFix);
  height = (font->Height() > CELLHEIGHT) ? font->Height() : CELLHEIGHT;
  
  if (!config.OverrideColors) {
    bgClr = clrBlack;
    fgClr = clrWhite;
  }
  else
  {
    bgClr = colors[config.BgClr];
    fgClr = colors[config.FgClr];
  }  
  
  cRow = 15;
  cCol = 0;
  mode = M_NONE;
}


//-----------------------------------------------------------------------------

CCOsd::~CCOsd()
{
  osdMutex->Lock();
  delete osd;
  osd = NULL;
  osdMutex->Unlock();
  
  scrollerMutex->Lock();
  delete scroller;
  scroller = NULL;
  scrollerMutex->Unlock();
  
  bufferMutex->Lock();
  delete offScreenBuffer;
  offScreenBuffer = NULL;
  bufferMutex->Unlock();
    
  delete osdMutex;
  delete scrollerMutex;
  delete bufferMutex;
}


//-----------------------------------------------------------------------------

void CCOsd::Hide()
{
  cMutexLock(osdMutex);
  delete osd;
  osd = NULL;
}


//-----------------------------------------------------------------------------

void CCOsd::Show()
{
  cMutexLock(osdMutex);
  if (osd) {     
    dprint(L_ERR, "CCOsd   : Show() OSD is not NULL.");
    return;
  }
  
  int width = (OSDWIDTH - (OSDWIDTH & 3))+3; // Make sure we don't get 
                                             // a wrong alignment error
  tArea Area = { 0, 0, width, OSDHEIGHT, 2 };
 
  osd = cOsdProvider::NewOsd(STARTX, STARTY, true);
      
  eOsdError result = osd->CanHandleAreas(&Area, 1);
  if (result == oeOk)
  {      
    osd->SetAreas(&Area, 1);
      
    dprint(L_DBG, "CCOsd   : Created (%d x %d)", width, OSDHEIGHT);
      
    osd->DrawRectangle(0,0,OSDWIDTH,OSDHEIGHT,clrTransparent);
    osd->Flush();
    
    cMutexLock(bufferMutex);
    offScreenBuffer->DrawRectangle(0,0,OSDWIDTH,OSDHEIGHT,clrTransparent);
    
    //DrawGrid();
    osd->Flush();  
  }
  else 
  { 
    dprint(L_ERR, "CCOsd   : Error (#%d) %s", result, OsdErrorToString(result));
  }
}


//-----------------------------------------------------------------------------
/*
void CCOsd::DrawGrid(void) 
{
  for (int x=0; x<=OSDWIDTH; x+= CELLWIDTH)
    osd->DrawRectangle(x,0 ,x,OSDHEIGHT, clrWhite);
  for (int y=0; y<=OSDHEIGHT; y+= CELLHEIGHT)
    osd->DrawRectangle(0,y ,OSDWIDTH,y, clrWhite);
}
*/

//-----------------------------------------------------------------------------

void CCOsd::ClearScreen(void)
{
  cMutexLock(osdMutex);

  if (osd) {
    osd->DrawRectangle(0,0,OSDWIDTH,OSDHEIGHT,clrTransparent);
    osd->Flush();
  }
}


//-----------------------------------------------------------------------------

void CCOsd::ClearBuffer(void)
{ 
  cMutexLock(bufferMutex);
  offScreenBuffer->DrawRectangle(0,0,OSDWIDTH,OSDHEIGHT,clrTransparent);
}


//-----------------------------------------------------------------------------

void CCOsd::Flush(void)
{
  cMutexLock(osdMutex);
  if (osd) {
    osd->DrawBitmap(0,0, *offScreenBuffer);
    osd->Flush();
  }
  ClearBuffer();  
}


//-----------------------------------------------------------------------------

void CCOsd::SetPosition(int row, int column)
{
  cRow = row;  cCol = column;
}


//-----------------------------------------------------------------------------

bool CCOsd::SetMode(captionMode m, int arg) 
{ 
  // dprint(L_DBG, "CCOsd   : Mode change %d -> %d", mode, m);
  if (mode == m) return false; // return false if we don't change modes
  mode = m;
  
  cMutexLock(scrollerMutex);
  
  if (mode == M_RU) {
    if (!scroller) scroller = new Scroller(arg);
  }  
  else {
    delete scroller;
    scroller = NULL;
  }
  
  return true;   
}


//-----------------------------------------------------------------------------

void CCOsd::AddChar(uint8_t c)
{
  if (mode == M_RCL) {
    cMutexLock(bufferMutex);
    AddCharToBuffer(c);
  }  
  else if (mode == M_RDC) {
    cMutexLock(osdMutex);
    AddCharToScreen(c);
  }  
  else if (mode == M_RU ) 
  {
    cMutexLock(scrollerMutex);
    if (scroller) {
      AddCharToScreen(c); 
      scroller->AddChar(c);
    }   
  }   
}


//-----------------------------------------------------------------------------
#include "bitmaps.h"

void CCOsd::AddCharToBuffer(uint8_t c)
{ 
  int x = cCol*CELLWIDTH;
  int y = cRow*CELLHEIGHT;
  
  if (c == 0x01) { // Note
    DrawBitmapToBuffer(x,y, noteBitmap);
  }  
  else if (c != 0x00) // Use 0x00 as transparent space
  {
    char str[2] = { c, '\0' };
    offScreenBuffer->DrawText(x, y, str, fgClr, bgClr, font, CELLWIDTH, height, taCenter);
  }  
  cCol++;
}


//-----------------------------------------------------------------------------

void CCOsd::AddCharToScreen(uint8_t c)
{
  if (!osd) return;
  
  int x = cCol*CELLWIDTH;
  int y = cRow*CELLHEIGHT;
  
  if (c == 0x01) { // Note
    DrawBitmapToScreen(x,y, noteBitmap);
  }
  else if (c != 0x00) // Use 0x00 as transparent space
  {
    char str[2] = { (char) c, '\0' };
    osd->DrawText(x, y, str, fgClr, bgClr, font, CELLWIDTH, height, taCenter);
    osd->Flush();
  }
  cCol++;
  
}


//-----------------------------------------------------------------------------

void CCOsd::DrawBitmapToBuffer(int x, int y, const bool bitmap[24][18])
{
  for (int r=0; r<24; r++)
    for (int c=0; c<18; c++)
      offScreenBuffer->DrawPixel(x+c,y+r, bitmap[r][c] ? fgClr : bgClr); 
}


//-----------------------------------------------------------------------------

void CCOsd::DrawBitmapToScreen(int x, int y, const bool bitmap[24][18])
{
  for (int r=0; r<24; r++)
    for (int c=0; c<18; c++)
      osd->DrawPixel(x+c,y+r, bitmap[r][c] ? fgClr : bgClr); 
      
  osd->Flush();    
}


//-----------------------------------------------------------------------------

void CCOsd::Backspace(void)
{ 
  cCol--;
  if (mode == M_RCL) {
    cMutexLock(bufferMutex);
    offScreenBuffer->DrawRectangle(cCol*CELLWIDTH, cRow*CELLHEIGHT, CELLWIDTH, height, clrTransparent);
  }
  else if (mode == M_RDC) 
  {
    cMutexLock(osdMutex);
    if (osd) {
      osd->DrawRectangle(cCol*CELLWIDTH, cRow*CELLHEIGHT, CELLWIDTH, height, clrTransparent);
      osd->Flush();
    }
  }
  else if (mode == M_RU) {
    cMutexLock(scrollerMutex);
    if (scroller)
      scroller->Backspace();
  }
}


//-----------------------------------------------------------------------------

void CCOsd::CR(void)
{ 
  cMutexLock(scrollerMutex);
   
  if (mode == M_RU && scroller) 
  {
    ClearScreen();
    
    cMutexLock(osdMutex);
    
    scroller->Scroll();
    if (!osd) return;
    
    int lines = scroller->Lines();
    for (int r=0; r<lines-1; r++)
      for (int c=0; c<NUMCOL; c++)
      {
        int x = (c+0)*CELLWIDTH;
        int y = (cRow-lines+r+1)*CELLHEIGHT;
        
        char chr = scroller->Get(r, c);
        if (chr != '\0') {
          char str[2] = { chr, '\0' };
          osd->DrawText(x, y, str, fgClr, bgClr, font, CELLWIDTH, height, taCenter);
        }
      }
    osd->Flush();
  }
}


//-----------------------------------------------------------------------------

void CCOsd::DER(void)
{ 
  if (mode == M_RCL) {
    cMutexLock(bufferMutex);
    offScreenBuffer->DrawRectangle(cCol*CELLWIDTH, cRow*CELLHEIGHT, OSDWIDTH-cCol*CELLWIDTH, height, clrTransparent);
  }
  else if (mode == M_RDC) {
    cMutexLock(osdMutex);
    if (osd) {
      osd->DrawRectangle(cCol*CELLWIDTH, cRow*CELLHEIGHT, OSDWIDTH-cCol*CELLWIDTH, height, clrTransparent);
      osd->Flush();
    }
  }
  else if (mode == M_RU) {
    cMutexLock(scrollerMutex);
    if (scroller) {
      scroller->DER();
      // Not finished...
    } 
  }  
}


//-----------------------------------------------------------------------------

void CCOsd::Reset(void)
{
  ClearScreen();
  ClearBuffer();
  
  cMutexLock(scrollerMutex);
  if (scroller) {
    delete scroller;
    scroller = NULL;
  }
  
  mode = M_NONE;
}


///////////////////////////////////////////////////////////////////////////////

#include <string.h>


Scroller::Scroller(int lines)
{                
  dprint(L_DBG, "Scroller: Creating with %d lines", lines);
  data = new char[NUMCOL * lines];
  memset(data, '\0', NUMCOL * lines);
  numLines = lines;
  cCol=0;
}


//-----------------------------------------------------------------------------

Scroller::~Scroller()
{
  delete[] data;
}


//-----------------------------------------------------------------------------

char Scroller::Get(int r, int c) 
{ 
  if (r<numLines && c<NUMCOL)
    return data[r*NUMCOL+c];
  else
    return '\0';   
}


//-----------------------------------------------------------------------------

void Scroller::Put(int r, int c, char ch) 
{  
  if (r<numLines && c<NUMCOL) {
    data[r*NUMCOL+c] = ch; 
  } 
}


//-----------------------------------------------------------------------------

void Scroller::AddChar(char c)
{
  Put(numLines-1, cCol, c);
  cCol++;
}
 
//-----------------------------------------------------------------------------
 
void Scroller::Scroll(void)
{
  for (int i=0; i<numLines-1; i++) {
    memcpy(data+i*NUMCOL, data+(i+1)*NUMCOL, NUMCOL);
  }  
    
  memset(data+(numLines-1)*NUMCOL, '\0', NUMCOL);
  cCol=0;
}


//-----------------------------------------------------------------------------

void Scroller::Backspace(void)
{
  cCol--;
  Put(numLines-1, cCol, '\0');
}


//-----------------------------------------------------------------------------

void Scroller::DER(void)
{
 // Untested
  memset(data+(numLines-1)*NUMCOL+cCol, '\0', NUMCOL-cCol); 
}


//-----------------------------------------------------------------------------

void Scroller::Reset(void)
{
  memset(data, '\0', NUMCOL * numLines);
  cCol=0;
}


//-----------------------------------------------------------------------------

void Scroller::Print(void)
{
  fprintf(stderr, "\n");
  for (int r=0; r<numLines; r++) {
    fprintf(stderr, "%d [ ", r);
    for (int c=0; c<NUMCOL; c++)
      fprintf(stderr, "%c", data[r*NUMCOL+c]);
    fprintf(stderr, " ]\n");  
  }
}


///////////////////////////////////////////////////////////////////////////////

 
